/*
Copyright Ant Financial. All Rights Reserved.

SPDX-License-Identifier: Apache-2.0
*/

package main

import (
	"encoding/base64"
	"encoding/json"
	"fmt"
	"github.com/golang/protobuf/proto"
	"github.com/hyperledger/fabric-chaincode-go/shim"
	"github.com/hyperledger/fabric-protos-go/common"
	"github.com/hyperledger/fabric-protos-go/msp"
	pb "github.com/hyperledger/fabric-protos-go/peer"
	"github.com/pkg/errors"
	"math"
)

func main() {
	err := shim.Start(new(ConsortiumManagementChaincode))
	if err != nil {
		fmt.Printf("Error starting consortium management chaincode: %s", err)
	}
}

// ConsortiumManagementChaincode serves functionalities to communicate channel updates and signatures between different channel members.
type ConsortiumManagementChaincode struct {
}

// Proposal is the main data structure used in CMCC interface to describe a proposal in Fabric blockchain network.
// proposal_type has the following string enumeration types, and for unlisted or unsupported types, the proposal should be ignored:
//	 ConfigUpdate: proposal_data is base64-encoded string representation of common.ConfigUpdate; the value of signatures is base64-encoded string representation of common.ConfigSignature.
type Proposal struct {
	// contains the msp ID of the proposal creator
	Creator string `json:"creator"`

	// define the type of the proposal, see section proposal_type
	Type string `json:"proposal_type"`

	// the content of the proposal, the type of data stored is determined by proposal_type
	Data string `json:"proposal_data"`

	// describes the proposal
	Description string `json:"description"`

	// the key of map is the msp ID of organization, the value of map is the signature data and determined by proposal_type
	Signatures map[string]string `json:"signatures"`

	// the time that proposal created in unix timestamp
	CreateTime int64 `json:"create_time,omitempty"`
}

const (
	NewProposalEvent    = "newProposalEvent"
	DeleteProposalEvent = "deleteProposalEvent"
	NewSignatureEvent   = "newSignatureEvent"
)

const (
	Proposal_Type_ConfigUpdate        = "ConfigUpdate"
	Proposal_Type_ChaincodeDefinition = "ChaincodeDefinition"
)

type ChaincodeDefinition struct {
	// Chaincode name
	Name string `json:"name"`

	// Chaincode verison
	Version string `json:"version"`

	// Chaincode package hash or id
	PkgId string `json:"pkg_id"`

	// Chaincode whether needs init, only available in lifecycle 2.0
	InitRequired bool `json:"init_required,omitempty"`

	// Chaincode endorsement policy, channel config policy MUST start with '/', else it is a signature policy, only available in lifecycle 2.0
	EndorsementPolicy string `json:"endorsement_policy,omitempty"`

	// Chaincode definition sequence number, only available in lifecycle 2.0
	Sequence int64 `json:"sequence,omitempty"`

	// Chaincode collection config, only available in lifecycle 2.0
	CollectionConfig string `json:"collection_config,omitempty"`

	// Chaincode download url address
	PkgUrl string `json:"pkg_url,omitempty"`
}

func (c *ConsortiumManagementChaincode) Init(stub shim.ChaincodeStubInterface) pb.Response {
	return shim.Success(nil)
}

func (c *ConsortiumManagementChaincode) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	function, args := stub.GetFunctionAndParameters()
	switch function {
	case "proposeConfigUpdate":
		return c.proposeConfigUpdate(stub, args)
	case "proposeChaincodeDefinition":
		return c.proposeChaincodeDefinition(stub, args)
	case "addSignature":
		return c.addSignature(stub, args)
	case "getProposals":
		return c.getProposals(stub, args)
	case "getProposal":
		return c.getProposal(stub, args)
	case "deleteProposal":
		return c.deleteProposal(stub, args)
	case "version":
		return shim.Success([]byte("v1.5"))
	default:
		return shim.Error("Code 400: Invalid invoke function name. Expecting \"proposeConfigUpdate\" \"proposeChaincodeDefinition\" \"addSignature\" \"getProposals\" \"getProposal\" \"deleteProposal\".")
	}
}

// proposeConfigUpdate proposes a channel configuration update proposal, and the organizations in the channel can sign(vote) on the proposal
//
// Arguments:
// 	1: proposalID 	- The unique ID of the proposal, a string of up to 64 characters [a-z][A-Z][-_], cannot be the same as any existing proposal ID
// 	2: configUpdate - The content of channel config update, base64-encoded string representation of common.ConfigUpdate
// 	3: description	- It describes the proposal
//
// Returns:
//   the ID of the created proposal
//
// Events:
//   name: newProposalEvent
//   payload: ID of the proposal
//
func (c *ConsortiumManagementChaincode) proposeConfigUpdate(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 3 {
		return shim.Error("Code 400: incorrect number of arguments - expecting 3: proposalID, configUpdate, description")
	}

	proposalID := args[0]
	configUpdate := args[1]
	description := args[2]

	// check configUpdate, base64 encoded common.ConfigUpdate
	update, err := base64.StdEncoding.DecodeString(configUpdate)
	if err != nil {
		return shim.Error("Code 400: base64 decode configUpdate failed: " + err.Error())
	}
	if err := proto.Unmarshal(update, &common.ConfigUpdate{}); err != nil {
		return shim.Error("Code 400: proto unmarshal common.ConfigUpdate failed: " + err.Error())
	}

	// check id is valid
	_, _, err = getProposal(stub, proposalID)
	if err == nil {
		return shim.Error("Code 400: ProposalID already in use.")
	} else {
		if _, ok := err.(ErrProposalNotFound); !ok {
			return shim.Error("Code 500: " + err.Error())
		}
	}

	// get msp id from proposal
	mspID, err := getMSPID(stub)
	if err != nil {
		return shim.Error("Code 500: " + err.Error())
	}

	// create and store the proposal
	proposal := Proposal{
		Creator:     mspID,
		Type:        Proposal_Type_ConfigUpdate,
		Data:        configUpdate,
		Description: description,
	}
	propsosalJSON, err := json.Marshal(proposal)
	if err != nil {
		return shim.Error("Code 500: json marshal proposal failed: " + err.Error())
	}
	if err := stub.PutState(string(proposalID), propsosalJSON); err != nil {
		return shim.Error("Code 500: put state failed: " + err.Error())
	}
	if err = stub.SetEvent(NewProposalEvent, []byte(proposalID)); err != nil {
		return shim.Error("Code 500: emit event failed: " + err.Error())
	}
	return shim.Success([]byte(fmt.Sprintf("{\"proposal_id\":\"%v\"}", proposalID)))
}

// proposeChaincodeDefinition saves a chaincode definition
//
// Arguments:
// 	1: proposalID 	- The unique ID of the proposal, a string of up to 64 characters [a-z][A-Z][-_], cannot be the same as any existing proposal ID
// 	2: definition   - The chaincode definition
// 	3: description	- It describes the proposal
//
// Returns:
//   the ID of the created proposal
//
// Events:
//   name: newProposalEvent
//   payload: ID of the proposal
//
func (c *ConsortiumManagementChaincode) proposeChaincodeDefinition(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 3 {
		return shim.Error("Code 400: incorrect number of arguments - expecting 3: proposalID, configUpdate, description")
	}

	proposalID := args[0]
	definition := args[1]
	description := args[2]
	def := &ChaincodeDefinition{}

	// check chaincode definition
	definitionBytes, err := base64.StdEncoding.DecodeString(definition)
	if err != nil {
		return shim.Error("Code 400: base64 decode chaincode definition failed: " + err.Error())
	}
	err = json.Unmarshal([]byte(definitionBytes), def)
	if err != nil {
		return shim.Error("Code 400: json unmarshal chaincode definition failed: " + err.Error())
	}
	if def.Name == "" || def.Version == "" || def.PkgId == "" {
		return shim.Error("Code 400: invalid chaincode definition")
	}

	// check id is valid
	_, _, err = getProposal(stub, proposalID)
	if err == nil {
		return shim.Error("Code 400: ProposalID already in use.")
	} else {
		if _, ok := err.(ErrProposalNotFound); !ok {
			return shim.Error("Code 500: " + err.Error())
		}
	}

	// get msp id from proposal
	mspID, err := getMSPID(stub)
	if err != nil {
		return shim.Error("Code 500: " + err.Error())
	}

	// create and store the proposal
	proposal := Proposal{
		Creator:     mspID,
		Type:        Proposal_Type_ChaincodeDefinition,
		Data:        definition,
		Description: description,
	}
	propsosalJSON, err := json.Marshal(proposal)
	if err != nil {
		return shim.Error("Code 500: json marshal proposal failed: " + err.Error())
	}
	if err := stub.PutState(string(proposalID), propsosalJSON); err != nil {
		return shim.Error("Code 500: put state failed: " + err.Error())
	}
	if err = stub.SetEvent(NewProposalEvent, []byte(proposalID)); err != nil {
		return shim.Error("Code 500: emit event failed: " + err.Error())
	}
	return shim.Success([]byte(fmt.Sprintf("{\"proposal_id\":\"%v\"}", proposalID)))
}

// addSignature adds the signature of an organization to an existing proposal.
//
// Arguments:
//   0: proposalID - the ID of the proposal where the signature is added
//   1: signature  - base64 encoded common.ConfigSignature
//
// Events:
//   name: newSignatureEvent
//   payload: ID of the proposal
//
func (c *ConsortiumManagementChaincode) addSignature(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 2 {
		return shim.Error("Code 400: incorrect number of arguments - expecting 2: proposalID, signature")
	}
	proposalID := args[0]
	signature := args[1]

	// check ConfigSignature is valid: base64 encoded common.ConfigSignature
	sig, err := base64.StdEncoding.DecodeString(signature)
	if err != nil {
		return shim.Error("Code 400: base64 deocode configUpdate failed: " + err.Error())
	}
	if err := proto.Unmarshal(sig, &common.ConfigSignature{}); err != nil {
		return shim.Error("Code 400: proto unmarshal common.ConfigSignature failed: " + err.Error())
	}

	// get msp id from proposal
	mspID, err := getMSPID(stub)
	if err != nil {
		return shim.Error("Code 500: " + err.Error())
	}

	// fetch and update the state of the proposal
	proposal, _, err := getProposal(stub, proposalID)
	if err != nil {
		if _, ok := errors.Cause(err).(ErrProposalNotFound); !ok {
			return shim.Error("Code 500: " + err.Error())
		}
		return shim.Error(fmt.Sprintf("Code 404: Proposal %s not found", proposalID))
	}
	if proposal.Signatures == nil {
		proposal.Signatures = make(map[string]string)
	}
	proposal.Signatures[mspID] = signature

	// store the updated proposal
	proposalJSONUpdated, err := json.Marshal(proposal)
	if err != nil {
		return shim.Error("Code 500: json marshal proposal failed: " + err.Error())
	}
	if err := stub.PutState(proposalID, proposalJSONUpdated); err != nil {
		return shim.Error("Code 500: put state failed: " + err.Error())
	}
	if err = stub.SetEvent(NewSignatureEvent, []byte(proposalID)); err != nil {
		return shim.Error("Code 500: emit event failed: " + err.Error())
	}
	return shim.Success([]byte(fmt.Sprintf("{\"proposal_id\":\"%v\"}", proposalID)))
}

// deleteProposal is used to delete an existing proposal in CMCC.
//
// Arguments:
//   0: proposalID - proposal id
//
// Events:
//   name: deleteProposalEvent
//   payload: ID of the proposal
//
func (c *ConsortiumManagementChaincode) deleteProposal(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 1 {
		return shim.Error("Code 400: incorrect number of arguments - expecting 1: proposalID")
	}
	proposalID := args[0]

	// fetch proposal
	proposal, proposalJSON, err := getProposal(stub, proposalID)
	if err != nil {
		if _, ok := errors.Cause(err).(ErrProposalNotFound); !ok {
			return shim.Error("Code 500: " + err.Error())
		}
		return shim.Error(fmt.Sprintf("Code 404: Proposal %s not found", proposalID))
	}

	mspID, err := getMSPID(stub)
	if err != nil {
		return shim.Error("Code 500: " + err.Error())
	}

	// check if calling organization is proposal creator
	if proposal.Creator != mspID {
		return shim.Error(fmt.Sprintf("Code 403: forbidden. only the proposal creator %s can delete the proposal", proposal.Creator))
	}

	// delete the proposal
	if err := stub.DelState(proposalID); err != nil {
		return shim.Error("Code 500: delete state failed: " + err.Error())
	}
	if err = stub.SetEvent(DeleteProposalEvent, []byte(proposalID)); err != nil {
		return shim.Error("Code 500: emit event failed: " + err.Error())
	}
	return shim.Success(proposalJSON)
}

// getProposals returns all proposals.
//
// Returns:
//   a map from proposalID to proposal
//
func (c *ConsortiumManagementChaincode) getProposals(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 0 {
		return shim.Error("Code 400: incorrect number of arguments - expecting 0")
	}
	proposals := make(map[string]*Proposal)
	proposalIterator, err := stub.GetStateByRange("", "")
	if err != nil {
		return shim.Error("Code 500: get state by range failed: " + err.Error())
	}
	defer proposalIterator.Close()

	for proposalIterator.HasNext() {
		proposalKV, err := proposalIterator.Next()
		if err != nil {
			return shim.Error("Code 500: iterating to next proposal failed: " + err.Error())
		}
		proposal := &Proposal{}
		if err = json.Unmarshal(proposalKV.Value, proposal); err != nil {
			return shim.Error(fmt.Sprintf("Code 500: json umarshal proposal %s failed: %s", proposalKV.Key, err.Error()))
		}
		createTime, _, err := getProposalCreateModifyTime(stub, proposalKV.Key)
		if err != nil {
			fmt.Printf("Get proposal %s craete time failed: %s\n", proposalKV.Key, err)
		}
		proposal.CreateTime = createTime
		proposals[proposalKV.Key] = proposal
	}

	proposalsJSON, err := json.Marshal(proposals)
	if err != nil {
		return shim.Error("Code 500: json marshal proposals failed: " + err.Error())
	}
	return shim.Success(proposalsJSON)
}

// getProposal returns the proposal with the given ID.
//
// Arguments:
//   0: proposalID - the ID of a proposal
//
// Returns:
//   the proposal with the given ID
//
func (c *ConsortiumManagementChaincode) getProposal(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 1 {
		return shim.Error("Code 400: incorrect number of arguments - expecting 1: proposalID")
	}
	proposalID := args[0]
	_, proposalJSON, err := getProposal(stub, proposalID)
	if err != nil {
		if _, ok := err.(ErrProposalNotFound); ok {
			return shim.Error("Code 404: " + err.Error())
		}
		return shim.Error("Code 500: " + err.Error())
	}
	return shim.Success(proposalJSON)
}

// get msp id from proposal
func getMSPID(stub shim.ChaincodeStubInterface) (string, error) {
	creator, err := stub.GetCreator()
	if err != nil {
		return "", errors.Wrap(err, "Get creator from stub failed")
	}

	identity := &msp.SerializedIdentity{}
	if err := proto.Unmarshal(creator, identity); err != nil {
		return "", errors.Wrap(err, "Proto unmarshal SerializedIdentity failed")
	}
	return identity.Mspid, nil
}

// ErrProposalNotFound is returned when the requested object is not found.
type ErrProposalNotFound string

func (e ErrProposalNotFound) Error() string {
	return fmt.Sprintf("proposal %s not found", string(e))
}

func getProposalCreateModifyTime(stub shim.ChaincodeStubInterface, id string) (createTime int64, modifyTime int64, err error) {
	iter, err := stub.GetHistoryForKey(id)
	if err != nil {
		return 0, 0, errors.Wrapf(err, "Get History failed for key %s", id)
	}
	createTime = math.MaxInt64
	for iter.HasNext() {
		data, err := iter.Next()
		if err != nil {
			return 0, 0, errors.Wrapf(err, "Iter to next history failed for key %s", id)
		}
		if createTime > data.Timestamp.Seconds {
			createTime = data.Timestamp.Seconds
		}
		if modifyTime < data.Timestamp.Seconds {
			modifyTime = data.Timestamp.Seconds
		}
	}
	return createTime, modifyTime, nil
}

// getProposal fetches and decodes the proposal with the given id from the state or returns an error.
func getProposal(stub shim.ChaincodeStubInterface, id string) (*Proposal, []byte, error) {
	proposalJSON, err := stub.GetState(id)
	if err != nil {
		return nil, nil, errors.Wrapf(err, "Get state failed with key %s", id)
	}
	if len(proposalJSON) == 0 {
		return nil, nil, ErrProposalNotFound(id)
	}
	proposal := &Proposal{}
	if err := json.Unmarshal(proposalJSON, proposal); err != nil {
		return nil, nil, errors.Wrapf(err, "Json unmarshal proposal %s failed", id)
	}
	createTime, _, err := getProposalCreateModifyTime(stub, id)
	if err != nil {
		fmt.Printf("Get proposal %s craete time failed: %s\n", id, err)
	}
	proposal.CreateTime = createTime
	proposalJSON, err = json.Marshal(proposal)
	if err != nil {
		return nil, nil, errors.Wrapf(err, "Json marshal proposal %s failed", id)
	}
	return proposal, proposalJSON, nil
}
